/*-
 * Copyright (c) 2005-2006 The FreeBSD Project
 * All rights reserved.
 *
 * Author: Shteryana Shopova <syrinx@FreeBSD.org>
 *
 * Redistribution of this software and documentation and use in source and
 * binary forms, with or without modification, are permitted provided that
 * the following conditions are met:
 *
 * 1. Redistributions of source code or documentation must retain the above
 *    copyright notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * a tool for sending SNMP get requests (GET, GETBULK, GETNEXT ) 
 */
#include <sys/queue.h>
#include <sys/types.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include <bsnmp/asn1.h>
#include <bsnmp/snmp.h>
#include <bsnmp/snmpclient.h>
#include "bsnmptc.h"
#include "bsnmptools.h"

const char helptxt[] =\
"\
bsnmpget [-adehn] [-b buffersize] [-I options] [-i filelist] [-l filename]\n\
	 [-M max-repetitions] [-N non-repeaters] [-o output] [-p pdu]\n\
	 [-r retries] [-s [trans::][community@][server][:port]] \n\
         [-t timeout] [-v version] OID [OID [OID [...]]]\n\
options: \n\
 -a		ingore syntax/access type and add OID to pdu\n\
 -d		debug - dump raw Pdus\n\
 -e		if error is returned in responce pdu, resend request\n\
		removing the variable that caused the error (use with\n\
		Get/GetNext)\n\
 -h		print this text\n\
 -n		interpret the input as numerical OIDs rather than parse them\n\
 -b buffersize	size of receive/transmit buffer\n\
 -I options	include each of the files in the list - options are (keep\n\
		order):\n\
	cut=OID		specify an initial OID that was cut from the file to\n\
			be appended\n\
	path=pathname	specify a path where to read the files from\n\
	file=filelist	a list of comma separated files, to which the two\n\
			options above will apply\n\
 -i filelist	comma-separated filelist to read string to oid mappings\n\
 -l filename	name of the client's socket if local sockets are used\n\
 -M max-repetitions	specify a value for max-repetitions (use with GetBulk\n\
			only - default:1)\n\
 -N non-repeaters	specify a value for non-repeaters (use with GetBulk\n\
			only)\n\
 -o output	specify result output : short | verbose | quiet (default:\n\
		short)\n\
 -p pdu		pdu type to send : get | getbulk | getnext (default: getbulk)\n\
 -r retries	number of retries when sending a request (default: 3)\n\
 -s [trans::][community@][server][:port]	server specification\n\
	trans		transport type: udp | stream | dgram (default: udp)\n\
	community	community name (default: public)\n\
	server		SNMP server name or IP address\n\
	port		server port (default: 161)\n\
 -t timeout	period to wait before a request times out (default: 3 sec)\n\
 -v version	SNMP version to use : 1 | 2 (default: 2)\n\
OID		object identifier\n\
";

int32_t debug_on = 0;

/* The snmp client structure from libbsnmp. */
struct snmp_client snmp_client;
struct snmp_toolinfo snmptool;

#define	DEBUG_SNMPGET	if (debug_on) fprintf

#define	PDUTYPE_BITS	0xf00		/* bits 8-11 for pdu type */
#define	MAXREP_BITS	0xff0000	/* bits 16-23 for max-repetitions
					   value */
#define	NONREP_BITS	0xff000000	/* bits 24-31 for non-repeaters value */

#define	SET_PDUTYPE(type)	(snmptool.flags |= ((type) << 8))
#define	GET_PDUTYPE		((snmptool.flags & PDUTYPE_BITS) >> 8)

#define	SET_MAXREP(i)		((snmptool.flags |= ((i) << 16)))
#define	GET_MAXREP		((snmptool.flags & MAXREP_BITS) >> 16)

#define	SET_NONREP(i)		((snmptool.flags |= ((i) << 24)))
#define	GET_NONREP		((snmptool.flags & NONREP_BITS) >> 24)

static int32_t
parse_max_repetitions(char opt, char *opt_arg)
{
	uint32_t v;

	if (opt_arg == NULL) {
		warnx("Option %c requires an argument", opt);
		return (-1);
	}

	v = strtoul(opt_arg, (void *) NULL, 10);

	if (v > SNMP_MAX_BINDINGS) {
		warnx( "Max repetitions value too big - %d maximum allowed",
		    SNMP_MAX_BINDINGS);
		return (-1);
	}

	SET_MAXREP(v);
	return (2);
}

static int32_t
parse_non_repeaters(char opt, char *opt_arg)
{
	uint32_t v;

	if (opt_arg == NULL) {
		warnx("Option %c requires an argument", opt);
		return (-1);
	}

	v = strtoul(opt_arg, (void *) NULL, 10);

	if (v > SNMP_MAX_BINDINGS) {
		warnx("Non repeaters  value too big - %d maximum allowed",
		    SNMP_MAX_BINDINGS);
		return (-1);
	}

	SET_NONREP(v);
	return (2);
}

static int32_t
parse_pdu_type(char opt, char *opt_arg)
{
	if (opt_arg == NULL) {
		warnx("Option %c requires an argument", opt);
		return (-1);
	}

	if (strncasecmp(opt_arg, "getbulk", strlen("getbulk")) == 0)
		SET_PDUTYPE(SNMP_PDU_GETBULK);
	else if (strncasecmp(opt_arg, "getnext", strlen("getnext")) == 0)
		SET_PDUTYPE(SNMP_PDU_GETNEXT);
	else if (strncasecmp(opt_arg, "get", strlen("get")) == 0)
		SET_PDUTYPE(SNMP_PDU_GET);
	else {
		warnx( "PDU type not supported - %s", opt_arg);
		return (-1);
	}

	return (2);
}

static int32_t
snmpget_parse_options(int argc, char **argv)
{
	int32_t opt, count, optnum = 0;
	char opts[] = "ab:dehI:i:l:M:N:no:p:r:s:t:v:";

	while ((opt = getopt(argc, argv, opts)) != EOF) {
	    switch (opt) {
		case 'a':
			if ((count = parse_skip_access(optarg)) < 0)
				return (-1);
			break;
		case 'b':
			if ((count = parse_buflen(opt, optarg)) < 0)
				return (-1);
			break;
		case 'd':
			if ((count = parse_debug(optarg)) < 0)
				return (-1);
			break;
		case 'e':
			if ((count = parse_errors(optarg)) < 0)
				return (-1);
			break;
		case 'h':
			parse_help(optarg);
			return (-1);
		case 'I':
			if ((count = parse_include(opt, optarg)) < 0)
				return (-1);
			break;
		case 'i':
			if ((count = parse_file(opt, optarg)) < 0)
				return (-1);
			break;
		case 'l':
			if ((count = parse_local_path(opt, optarg)) < 0)
				return (-1);
			break;
		case 'M':
			if ((count = parse_max_repetitions(opt, optarg)) < 0)
				return (-1);
			break;
		case 'N':
			if ((count = parse_non_repeaters(opt, optarg)) < 0)
				return (-1);
			break;
		case 'n':
			if ((count = parse_num_oids(optarg)) < 0)
				return (-1);
			break;
		case 'o':
			if ((count = parse_output(opt, optarg)) < 0)
				return (-1);
			break;
		case 'p':
			if ((count = parse_pdu_type(opt, optarg)) < 0)
				return (-1);
			break;
		case 'r':
			if ((count = parse_retry(opt, optarg)) < 0)
				return (-1);
			break;
		case 's':
			if ((count = parse_server(opt, optarg)) < 0)
				return (-1);
			break;
		case 't':
			if ((count = parse_timeout(opt, optarg)) < 0)
				return (-1);
			break;
		case 'v':
			if ((count = parse_version(opt, optarg)) < 0)
				return (-1);
			break;
		default:
			fprintf(stderr,"Unknown option - %c\n", opt);
			fprintf(stderr, "%s", helptxt);
			return (-1);
	    }
	    optnum += count;
	}

	return (optnum);
}

/*
 * Read user input OID - one of following formats:
 * 1) 1.2.1.1.2.1.0 - that is if option numeric was giveni;
 * 2) string - in such case append .0 to the asn_oid subs;
 * 3) string.1 - no additional proccessing required in such case.
 */
static int32_t
snmpget_parse_stroid(struct snmp_object *obj, char *argv)
{
	char string[MAXSTR], *ptr;
	int32_t i = 0;
	struct asn_oid in_oid;

	ptr = argv;
	while (isalpha(*ptr) || *ptr == '_' || (i != 0 && isdigit(*ptr))) {
		ptr++;
		i++;
	}

	if (i <= 0 || i >= MAXSTR)
		return (-1);

	memset(&in_oid, 0, sizeof(struct asn_oid));
	if ((ptr = snmp_parse_suboid((argv + i), &in_oid)) == NULL) {
		warnx("Invalid OID - %s", argv);
		return (-1);
	}

	strlcpy(string, argv, i + 1);
	if (snmp_lookup_oidall(obj, string) < 0) {
		warnx("No entry for %s in mapping lists", string);
		return (-1);
	}

	if (in_oid.len > 0)
		asn_append_oid(&(obj->val.var), &in_oid);
	else if (*ptr == '[') {
		if ((ptr = snmp_parse_index(ptr + 1, obj)) == NULL)
			return (-1);
	} else if (obj->val.syntax > 0 && GET_PDUTYPE == SNMP_PDU_GET) {
		if (snmp_suboid_append(&(obj->val.var), (asn_subid_t) 0) < 0)
			return (-1);
	}

	if (*ptr != '\0') {
		warnx("Invalid character after OID - %c", *ptr);
		return (-1);
	}

	return (1);
}

static int32_t
snmpget_parse_oid(struct snmp_object *obj, char *argv)
{
	if (argv == NULL)
		return (-1);

	if (ISSET_NUMERIC) {
		if (snmp_parse_numoid(argv, &(obj->val.var)) < 0)
			return (-1);
	} else {
		if (snmpget_parse_stroid(obj, argv) < 0)
			return (-1);
	}

	return (1);
}

static int32_t
snmpget_verify_vbind(struct snmp_pdu *pdu, struct snmp_object *obj)
{
	if (pdu->version == SNMP_V1 && obj->val.syntax ==
	    SNMP_SYNTAX_COUNTER64) {
		warnx("64-bit counters are not supported in SNMPv1 PDU");
		return (-1);
	}

	if (ISSET_NUMERIC || pdu->type == SNMP_PDU_GETNEXT || pdu->type ==
	    SNMP_PDU_GETBULK)
		return (1);

	if (pdu->type == SNMP_PDU_GET && obj->val.syntax == 0) {
		warnx("Only leaf object values can be added to GET PDU");
		return (-1);
	}

	return (1);
}

static int32_t
snmpget_add_vbind(struct snmp_pdu *pdu, struct snmp_object *obj)
{
	if (pdu->nbindings > SNMP_MAX_BINDINGS) {
		warnx("Too many bindings in PDU");
		return (-1);
	}

	if (obj->error > 0) {
		DEBUG_SNMPGET(stderr,"Object already returned error - "
		    "skipping\n");
		return (0);
	}

	asn_append_oid(&(pdu->bindings[pdu->nbindings].var), &(obj->val.var));
	pdu->nbindings++;

	return (pdu->nbindings);
}

/*
 * In case of a getbulk PDU, the error_status and error_index fields are used by
 * libbsnmp to hold the values of the non-repeaters and max-repetitions fields
 * that are present only in the getbulk - so before sending the PDU make sure
 * these have correct values as well.
 */
static int32_t
snmp_fix_getbulk(struct snmp_pdu *pdu, uint32_t max_rep, uint32_t non_rep)
{
	if (pdu == NULL)
		return (-1);

	if (pdu->nbindings < non_rep)
		pdu->error_status = pdu->nbindings;
	else
		pdu->error_status = non_rep;

	if (max_rep > 0)
		pdu->error_index = max_rep;
	else
		pdu->error_index = 1;

	return (1);
}

/*
 * According to command line options prepare SNMP Get | GetNext | GetBulk PDU.
 * Wait for a responce and print it.
 */
int
main(int argc, char ** argv)
{
	int32_t oid_cnt, last_oid, opt_num;
	struct snmp_pdu req, resp;

	snmptool_init(&snmptool, helptxt);

	if ((opt_num = snmpget_parse_options(argc, argv)) < 0) {
		snmp_tool_freeall();
		exit(0);
	}

	oid_cnt = argc - opt_num - 1;
	if (oid_cnt == 0) {
		fprintf(stderr, "Provide at least one OID\n");
		snmp_tool_freeall();
		exit(0);
	}

	if (oid_cnt <= 0 || oid_cnt > SNMP_MAX_BINDINGS) {
		fprintf(stderr, "Too many OIDs - %d\n", oid_cnt);
		snmp_tool_freeall();
		exit(0);
	}

	if (snmp_import_all() < 0) {
		snmp_tool_freeall();
		exit(0);
	}

	for (last_oid = argc - 1; oid_cnt > 0; last_oid--, oid_cnt--)
		if ((snmp_object_add(snmpget_parse_oid, argv[last_oid])) < 0) {
			snmp_tool_freeall();
			fprintf(stderr, "Error parsing OID string %s\n",
			    argv[last_oid]);
			exit(0);
		}

	/*
	 * A simple sanity check - can not send GETBULK when using SNMPv1.
	 */

	if (snmp_client.version == SNMP_V1 && GET_PDUTYPE == SNMP_PDU_GETBULK) {
		fprintf(stderr, "Can not send GETBULK PDU with SNMPv1\n");
		snmp_tool_freeall();
		exit(0);
	}

	if (snmp_open(NULL, NULL, NULL, NULL)) {
		warnx("Failed to open snmp session - %s", strerror(errno));
		snmp_tool_freeall();
		exit(-1);
	}

	snmp_pdu_create(&req, GET_PDUTYPE);

	while ((snmp_pdu_add_bindings(snmpget_verify_vbind, snmpget_add_vbind,
	    &req)) > 0) {
		if (GET_PDUTYPE == SNMP_PDU_GETBULK)
			snmp_fix_getbulk(&req, GET_MAXREP, GET_NONREP);

		if (snmp_dialog(&req, &resp) == -1) {
			warnx("Snmp dialog - %s", strerror(errno));
			break;
		}

		if (snmp_parse_resp(&resp,&req) >= 0) {
			snmp_output_resp(&resp);
			break;
		}

		snmp_output_err_resp(&resp);
		if (GET_PDUTYPE == SNMP_PDU_GETBULK || !ISSET_RETRY)
			break;

		/*
		 * Loop through the object list and set object->error to the
		 * varbinding that caused the error.
		 */
		if (snmp_object_seterror(&(resp.bindings[resp.error_index - 1]),
		    resp.error_status) <= 0)
			break;

		fprintf(stderr, "Retrying...\n");
		snmp_pdu_free(&resp);
		snmp_pdu_create(&req, GET_PDUTYPE);
	}

	snmp_pdu_free(&resp);
	snmp_tool_freeall();
	snmp_close();

	exit(0);
}
