/*-
 * Copyright (c) 2005-2006 The FreeBSD Project
 * All rights reserved.
 *
 * Author: Shteryana Shopova <syrinx@FreeBSD.org>
 *
 * Redistribution of this software and documentation and use in source and
 * binary forms, with or without modification, are permitted provided that
 * the following conditions are met:
 *
 * 1. Redistributions of source code or documentation must retain the above
 *    copyright notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * a tool for sending SNMP SET requests
 */
#include <sys/queue.h>
#include <sys/types.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include <bsnmp/asn1.h>
#include <bsnmp/snmp.h>
#include <bsnmp/snmpclient.h>
#include "bsnmptc.h"
#include "bsnmptools.h"

const char helptxt[] = "\
bsnmpset [-adehn] [-b buffersize] [-I options] [-i filelist] [-l filename]\n\
	 [-o output] [-r retries] [-s [trans::][community@][server][:port]]\n\
	 [-t timeout] [-v version]\n\
	 OID=syntax:value [OID=syntax:value [OID=syntax:value[...]]]\n\
options: \n\
 -a		ingore syntax/access type and add OID to pdu\n\
 -d		debug - dump raw Pdus\n\
 -e		if error is returned in responce pdu, resend request removing\n\
		the variable that caused the error\n\
 -h		print this text\n\
 -n		interpret the input as numerical OIDs rather than parse them\n\
 -b buffersize	size of receive/transmit buffer\n\
 -I options	include each of the files in the list - options are (keep\n\
		order):\n\
	cut=OID		specify an initial OID that was cut from the file to\n\
			be appended\n\
	path=pathname	specify a path where to read the files from\n\
	file=filelist	a list of comma separated files, to which the two\n\
			options above will apply\n\
 -i filelist	comma-separated filelist to read string to oid mappings\n\
 -l filename	name of the client's socket if local sockets are used\n\
 -o output	specify result output : short | verbose | tabular | quiet\n\
		(default: short)\n\
 -r retries	number of retries when sending a request (default: 3)\n\
 -s [trans::][community@][server][:port]	server specification\n\
	trans		transport type: udp | stream | dgram (default: udp)\n\
	community	community name (default: public)\n\
	server		SNMP server name or IP address\n\
	port		server port (default: 161)\n\
 -t timeout	period to wait before a request times out (default:\n\
		3 sec)\n\
 -v version	SNMP version to use : 1 | 2 (default: 2)\n\
OID=syntax:value	object identifier = syntax:value - no white spaces,\n\
			where syntax is\n\
	syntax		one of int (integer), ip (ipaddress),\n\
			cnt32 (counter32), gauge (gauge), ticks (timeticks),\n\
			cnt64 (counter64), oct(octetstring).\n\
	value		value to be set (ip addresses in form u.u.u.u).\n\
";

int32_t debug_on = 0;

/* The snmp client structure from libbsnmp. */
struct snmp_client snmp_client;
struct snmp_toolinfo snmptool;

#define	DEBUG_SNMPSET	if (debug_on) fprintf

static int32_t
snmpset_parse_options(int argc, char **argv)
{
	int32_t opt, count, optnum = 0;
	char opts[] = "ab:dehI:i:l:no:r:s:t:v:";

	while ((opt = getopt(argc, argv, opts)) != EOF) {
	    switch (opt) {
		case 'a':
			if ((count = parse_skip_access(optarg)) < 0)
				return (-1);
			break;
		case 'b':
			if ((count = parse_buflen(opt, optarg)) < 0)
				return (-1);
			break;
		case 'd':
			if ((count = parse_debug(optarg)) < 0)
				return (-1);
			break;
		case 'e':
			if ((count = parse_errors(optarg)) < 0)
				return (-1);
			break;
		case 'h':
			parse_help(optarg);
			return (-1);
		case 'I':
			if ((count = parse_include(opt, optarg)) < 0)
				return (-1);
			break;
		case 'i':
			if ((count = parse_file(opt, optarg)) < 0)
				return (-1);
			break;
		case 'l':
			if ((count = parse_local_path(opt, optarg)) < 0)
				return (-1);
			break;
		case 'n':
			if ((count = parse_num_oids(optarg)) < 0)
				return (-1);
			break;
		case 'o':
			if ((count = parse_output(opt, optarg)) < 0)
				return (-1);
			break;
		case 'r':
			if ((count = parse_retry(opt, optarg)) < 0)
				return (-1);
			break;
		case 's':
			if ((count = parse_server(opt, optarg)) < 0)
				return (-1);
			break;
		case 't':
			if ((count = parse_timeout(opt, optarg)) < 0)
				return (-1);
			break;
		case 'v':
			if ((count = parse_version(opt, optarg)) < 0)
				return (-1);
			break;
		default:
			fprintf(stderr,"Unknown option - %c\n", opt);
			fprintf(stderr, "%s", helptxt);
			return (-1);
	    }
	    optnum += count;
	}

	return (optnum);
}

static int32_t
parse_int(struct snmp_value *value, char *val)
{
	char *endptr;
	int32_t v, saved_errno;

	saved_errno = errno;
	errno = 0;

	v = strtol(val, &endptr, 10);

	if (errno != 0) {
		warnx("Value %s not supported - %s", val, strerror(errno));
		errno = saved_errno;
		return (-1);
	}

	value->syntax = SNMP_SYNTAX_INTEGER;
	value->v.integer = v;
	errno = saved_errno;

	return (0);
}

/*
 * Here syntax may be one of SNMP_SYNTAX_COUNTER, SNMP_SYNTAX_GAUGE,
 * SNMP_SYNTAX_TIMETICKS.
 */
static int32_t
parse_uint(struct snmp_value *value, char *val)
{
	char *endptr;
	uint32_t v = 0;
	int32_t saved_errno;

	saved_errno = errno;
	errno = 0;

	v = strtoul(val, &endptr, 10);

	if (errno != 0) {
		warnx("Value %s not supported - %s", val, strerror(errno));
		errno = saved_errno;
		return (-1);
	}

	value->v.uint32 = v;
	errno = saved_errno;

	return (0);
}

static int32_t
parse_ticks(struct snmp_value *value, char *val)
{
	if (parse_uint(value, val) < 0)
		return (-1);

	value->syntax = SNMP_SYNTAX_TIMETICKS;
	return (0);
}

static int32_t
parse_gauge(struct snmp_value *value, char *val)
{
	if (parse_uint(value, val) < 0)
		return (-1);

	value->syntax = SNMP_SYNTAX_GAUGE;
	return (0);
}

static int32_t
parse_counter(struct snmp_value *value, char *val)
{
	if (parse_uint(value, val) < 0)
		return (-1);

	value->syntax = SNMP_SYNTAX_COUNTER;
	return (0);
}

static int32_t
parse_uint64(struct snmp_value *value, char *val)
{
	char *endptr;
	int32_t saved_errno;
	uint64_t v;

	saved_errno = errno;
	errno = 0;

	v = strtoull(val, &endptr, 10);

	if (errno != 0) {
		warnx("Value %s not supported - %s", val, strerror(errno));
		errno = saved_errno;
		return (-1);
	}

	value->syntax = SNMP_SYNTAX_COUNTER64;
	value->v.counter64 = v;
	errno = saved_errno;

	return (0);
}

static int32_t
parse_oid_numeric(struct snmp_value *value, char *val)
{
	char *endptr;
	int32_t saved_errno;
	asn_subid_t suboid;

	do {
		saved_errno = errno;
		errno = 0;
		suboid = strtoul(val, &endptr, 10);
		if (errno != 0) {
			warnx("Value %s not supported - %s", val,
			    strerror(errno));
			errno = saved_errno;
			return (-1);
		}
		errno = saved_errno;
		if ((asn_subid_t) suboid > ASN_MAXID) {
			warnx("Suboid %u > ASN_MAXID", suboid);
			return (-1);
		}
		if (snmp_suboid_append(&(value->v.oid), suboid) < 0)
			return (-1);
		val = endptr + 1;
	} while (*endptr == '.');

	if (*endptr != '\0')
		warnx("OID value %s not supported", val);

	value->syntax = SNMP_SYNTAX_OID;
	return (0);
}

static int32_t
parse_ip(struct snmp_value * value, char * val)
{
	uint32_t v;
	int32_t i;
	char *endptr, *str;

	str = val;
	for (i = 0; i < 4; i++) {
		v = strtoul(str, &endptr, 10);
		if (v > 0xff)
			return (-1);
		if (*endptr != '.' && *endptr != '\0' && i != 3)
			break;
		str = endptr + 1;
		value->v.ipaddress[i] = (uint8_t) v;
	}

	value->syntax = SNMP_SYNTAX_IPADDRESS;
	return (0);
}

/*
 * Allow OID leaf in both forms:
 * 1) 1.3.6.1.2... ->  in such case call directly the function reading raw OIDs;
 * 2) begemotSnmpdAgentFreeBSD -> lookup the ASN OID corresponding to that.
 */
static int32_t
parse_oid_string(struct snmp_value *value, char *string)
{
	struct snmp_object obj;

	if (isdigit(string[0]))
		return (parse_oid_numeric(value, string));

	memset(&obj, 0, sizeof(struct snmp_object));
	if (snmp_lookup_enumoid(&obj, string) < 0) {
		warnx("Unknown OID enum string - %s", string);
		return (-1);
	}

	asn_append_oid(&(value->v.oid), &(obj.val.var));
	return (1);
}

static int32_t
parse_int_string(struct snmp_object *object, char *val)
{
	int32_t	v;

	if (isdigit(val[0]))
		return ((parse_int(&(object->val), val)));

	if (object->info == NULL) {
		warnx("Unknown enumerated integer type - %s", val);
		return (-1);
	}
	if ((v = enum_number_lookup(object->info->snmp_enum, val)) < 0)
		warnx("Unknown enumerated integer type - %s", val);

	object->val.v.integer = v;
	return (1);
}

static int32_t
parse_syntax_val(struct snmp_value *value, enum snmp_syntax syntax, char *val)
{
	switch (syntax) {
		case SNMP_SYNTAX_INTEGER:
			return (parse_int(value, val));
		case SNMP_SYNTAX_IPADDRESS:
			return (parse_ip(value, val));
		case SNMP_SYNTAX_COUNTER:
			return (parse_counter(value, val));
		case SNMP_SYNTAX_GAUGE:
			return (parse_gauge(value, val));
		case SNMP_SYNTAX_TIMETICKS:
			return (parse_ticks(value, val));
		case SNMP_SYNTAX_COUNTER64:
			return (parse_uint64(value, val));
		case SNMP_SYNTAX_OCTETSTRING:
			return (snmp_tc2oct(SNMP_STRING, value, val));
		case SNMP_SYNTAX_OID:
			return (parse_oid_numeric(value, val));
		default:
			/* NOTREACHED */
			break;
	}

	return (-1);
}

/*
 * Parse a command line argument of type OID=syntax:value and fill in whatever 
 * fields can be derived from the input into snmp_value structure. Reads numeric
 * OIDs.
 */
static int32_t
parse_pair_numoid_val(char *str, struct snmp_value *snmp_val)
{
	int32_t cnt;
	char *ptr;
	enum snmp_syntax syntax;
	char oid_str[ASN_OIDSTRLEN];

	ptr = str;
	for (cnt = 0; cnt < ASN_OIDSTRLEN; cnt++)
		if (ptr[cnt] == '=')
			break;

	if (cnt >= ASN_OIDSTRLEN) {
		warnx("OID too long - %s", str);
		return (-1);
	}
	strlcpy(oid_str, ptr, (size_t) (cnt + 1));

	ptr = str + cnt + 1;
	for (cnt = 0; cnt < MAX_CMD_SYNTAX_LEN; cnt++)
		if(ptr[cnt] == ':')
			break;

	if (cnt >= MAX_CMD_SYNTAX_LEN) {
		warnx("Unknown syntax in OID - %s", str);
		return (-1);
	}

	if ((syntax = parse_syntax(ptr)) <= SNMP_SYNTAX_NULL) {
		warnx("Unknown syntax in OID - %s", ptr);
		return (-1);
	}

	ptr = ptr + cnt + 1;
	for (cnt = 0; cnt < MAX_OCTSTRING_LEN; cnt++)
		if (ptr[cnt] == '\0')
			break;

	if (ptr[cnt] != '\0') {
		warnx("Value string too long - %s",ptr);
		return (-1);
	}

	/*
	 * Here try parsing the OIDs and syntaxes and then check values - have
	 * to know syntax to check value boundaries.
	 */
	if (snmp_parse_numoid(oid_str, &(snmp_val->var)) < 0) {
		warnx("Error parsing OID %s",oid_str);
		return (-1);
	}

	if (parse_syntax_val(snmp_val, syntax, ptr) < 0)
		return (-1);

	return (1);
}

static int32_t
parse_syntax_strval(char *str, struct snmp_object *object)
{
	uint32_t len;
	enum snmp_syntax syn;

	/*
	 * Syntax string here not required  - still may be present.
	 */

	if (GET_OUTPUT == OUTPUT_VERBOSE) {
		for (len = 0 ; *(str + len) != ':'; len++) {
			if (*(str + len) == '\0') {
				warnx("Syntax missing in value - %s", str);
				return (-1);
			}
		}
		if ((syn = parse_syntax(str)) <= SNMP_SYNTAX_NULL) {
			warnx("Unknown syntax in - %s", str);
			return (-1);
		}
		if (syn != object->val.syntax) {
			if (!ISSET_ERRIGNORE) {
				warnx("Bad syntax in - %s", str);
				return (-1);
			} else
				object->val.syntax = syn;
		}
		len++;
	} else
		len = 0;

	switch (object->val.syntax) {
		case SNMP_SYNTAX_INTEGER:
			return (parse_int_string(object, str + len));
		case SNMP_SYNTAX_IPADDRESS:
			return (parse_ip(&(object->val), str + len));
		case SNMP_SYNTAX_COUNTER:
			return (parse_counter(&(object->val), str + len));
		case SNMP_SYNTAX_GAUGE:
			return (parse_gauge(&(object->val), str + len));
		case SNMP_SYNTAX_TIMETICKS:
			return (parse_ticks(&(object->val), str + len));
		case SNMP_SYNTAX_COUNTER64:
			return (parse_uint64(&(object->val), str + len));
		case SNMP_SYNTAX_OCTETSTRING:
			return (snmp_tc2oct(object->info->tc, &(object->val),
			    str + len));
		case SNMP_SYNTAX_OID:
			return (parse_oid_string(&(object->val), str + len));
		default:
			/* NOTREACHED */
			break;
	}

	return (-1);
}

static int32_t
parse_pair_stroid_val(struct snmp_object *obj, char *argv)
{
	int32_t i;
	char string[MAXSTR], *ptr;
	struct asn_oid in_oid;

	ptr = argv;
	i = 0;
	while (isalpha(*ptr)|| *ptr == '_' || (i != 0 && isdigit(*ptr))) {
		ptr++;
		i++;
	}

	if (i <= 0 || i >= MAXSTR)
		return (-1);

	memset(&in_oid, 0, sizeof(struct asn_oid));
	if ((ptr = snmp_parse_suboid(argv + i, &in_oid)) == NULL) {
		warnx("Invalid OID - %s", argv);
		return(-1);
	}

	strlcpy(string, argv, i + 1);
	if (snmp_lookup_oidall(obj, string) < 0) {
		warnx("No entry for %s in mapping lists", string);
		return (-1);
	}

	if (in_oid.len > 0)
		asn_append_oid(&(obj->val.var), &in_oid);
	else if (*ptr == '[') {
		if ((ptr = snmp_parse_index(ptr + 1, obj)) == NULL)
			return (-1);
	} else {
		if (snmp_suboid_append(&(obj->val.var), 0) < 0)
			return (-1);
	}
 
	if (*ptr != '=') {
		warnx("Value to set expected after OID");
		return (-1);
	}

	if (parse_syntax_strval(ptr + 1, obj) < 0)
		return (-1);

	return (1);
}

static int32_t
snmpset_parse_oid(struct snmp_object *obj, char *argv)
{
	if (argv == NULL)
		return (-1);

	if (ISSET_NUMERIC) {
		if (parse_pair_numoid_val(argv, &(obj->val)) < 0)
			return (-1);
	} else {
		if (parse_pair_stroid_val(obj, argv) < 0)
			return (-1);
	}

	return (1);
}

static int32_t
add_ip_syntax(struct snmp_value *dst, struct snmp_value *src)
{
	int8_t i;

	dst->syntax = SNMP_SYNTAX_IPADDRESS;
	for (i = 0; i < 4; i++)
		dst->v.ipaddress[i] = src->v.ipaddress[i];

	return (1);
}

static int32_t
add_octstring_syntax(struct snmp_value *dst, struct snmp_value *src)
{
	if (src->v.octetstring.len > ASN_MAXOCTETSTRING) {
		warnx("OctetString len too big - %u",src->v.octetstring.len);
		return (-1);
	}

	if ((dst->v.octetstring.octets = malloc(src->v.octetstring.len)) ==
	    NULL) {
		syslog(LOG_ERR, "malloc() failed - %s", strerror(errno));
		return (-1);
	}

	memcpy(dst->v.octetstring.octets, src->v.octetstring.octets,
	    src->v.octetstring.len);
	dst->syntax = SNMP_SYNTAX_OCTETSTRING;
	dst->v.octetstring.len = src->v.octetstring.len;
	return(0);
}

static int32_t
add_oid_syntax(struct snmp_value *dst, struct snmp_value *src)
{
	asn_append_oid(&(dst->v.oid), &(src->v.oid));
	dst->syntax = SNMP_SYNTAX_OID;
	return (0);
}

/*
 * Check syntax - if one of SNMP_SYNTAX_NULL, SNMP_SYNTAX_NOSUCHOBJECT,
 * SNMP_SYNTAX_NOSUCHINSTANCE, SNMP_SYNTAX_ENDOFMIBVIEW or anything not known -
 * return error.
 */
static int32_t
snmpset_add_value(struct snmp_value *dst,struct snmp_value *src)
{
	if (dst == NULL || src == NULL)
		return (-1);

	switch(src->syntax) {
		case SNMP_SYNTAX_INTEGER:
			dst->v.integer = src->v.integer;
			dst->syntax = SNMP_SYNTAX_INTEGER;
			break;
		case SNMP_SYNTAX_TIMETICKS:
			dst->v.uint32 = src->v.uint32;
			dst->syntax = SNMP_SYNTAX_TIMETICKS;
			break;
		case SNMP_SYNTAX_GAUGE:
			dst->v.uint32 = src->v.uint32;
			dst->syntax = SNMP_SYNTAX_GAUGE;
			break;
		case SNMP_SYNTAX_COUNTER:
			dst->v.uint32 = src->v.uint32;
			dst->syntax = SNMP_SYNTAX_COUNTER;
			break;
		case SNMP_SYNTAX_COUNTER64:
			dst->syntax = SNMP_SYNTAX_COUNTER64;
			dst->v.counter64 = src->v.counter64;
			break;
		case SNMP_SYNTAX_IPADDRESS:
			add_ip_syntax(dst, src);
			break;
		case SNMP_SYNTAX_OCTETSTRING:
			add_octstring_syntax(dst, src);
			break;
		case SNMP_SYNTAX_OID:
			add_oid_syntax(dst, src);
			break;
		default:
			warnx("unknown syntax %d", src->syntax);
			return (-1);
	}

	return (0);
}

static int32_t
snmpset_verify_vbind(struct snmp_pdu *pdu, struct snmp_object *obj)
{
	/*
	 * Do not allow a 64-bit counter to be added to a SNMPv1 PDU.
	 */
	if (pdu->version == SNMP_V1 && obj->val.syntax == SNMP_SYNTAX_COUNTER64)
		return (-1);

	if (ISSET_NUMERIC || ISSET_ERRIGNORE)
		return (1);

	if (obj->info->access < SNMP_ACCESS_SET) {
		warnx("Object %s not accessible for set - try 'bsnmpset -a'",
		    obj->info->string);
		return (-1);
	}

	return (1);
}

static int32_t
snmpset_add_vbind(struct snmp_pdu *pdu, struct snmp_object *obj)
{
	if (pdu->nbindings > SNMP_MAX_BINDINGS) {
		warnx("Too many OIDs for one PDU");
		return (-1);
	}

	if (obj->error > 0)
		return (0);

	if (snmpset_add_value(&(pdu->bindings[pdu->nbindings]), &(obj->val))
	    < 0)
		return (-1);

	asn_append_oid(&(pdu->bindings[pdu->nbindings].var), &(obj->val.var));
	pdu->nbindings++;

	return (pdu->nbindings);
}

/*
 * According to command line options prepare SNMP Set PDU.
 * Wait for a responce and print it.
 */
int
main(int argc, char **argv)
{
	int32_t oid_cnt, last_oid, opt_num;
	struct snmp_pdu req, resp;

	snmptool_init(&snmptool,helptxt);

	if ((opt_num = snmpset_parse_options(argc,argv)) < 0) {
		snmp_tool_freeall();
		exit(0);
	}

	oid_cnt = argc - opt_num - 1;
	if (oid_cnt == 0) {
		warnx("Provide at least one OID");
		snmp_tool_freeall();
		exit(0);
	}

	if ((oid_cnt < 0)||(oid_cnt > SNMP_MAX_BINDINGS)) {
		warnx("Too many OIDs - %d", oid_cnt);
		snmp_tool_freeall();
		exit(0);
	}

	if (snmp_import_all() < 0) {
		snmp_tool_freeall();
		exit(0);
	}

	/* Parse the input OIDs vice versa. */
	for (last_oid = argc - 1; oid_cnt > 0; last_oid--, oid_cnt--)
		if ((snmp_object_add(snmpset_parse_oid, argv[last_oid])) < 0) {
			snmp_tool_freeall();
			fprintf(stderr, "Error parsing OID string %s\n",
			    (argv[last_oid]));
			exit(0);
		}

	if (snmp_open(NULL, NULL, NULL, NULL)) {
		warnx("Failed to open snmp session - %s", strerror (errno));
		snmp_tool_freeall();
		exit(-1);
	}

	snmp_pdu_create(&req, SNMP_PDU_SET);
	while ((snmp_pdu_add_bindings(snmpset_verify_vbind, snmpset_add_vbind,
	    &req)) > 0) {
		if(snmp_dialog(&req, &resp)) {
			warnx("Snmp dialog - %s", strerror (errno));
			break;
		}

		if (snmp_pdu_check(&req, &resp) > 0) {
			if (GET_OUTPUT != OUTPUT_QUIET)
				snmp_output_resp(&resp);
			break;
		}

		snmp_output_err_resp(&resp);
		if (!ISSET_RETRY)
			break;

		if (snmp_object_seterror(&(resp.bindings[resp.error_index - 1]),
		    resp.error_status) <= 0)
			break;
		fprintf(stderr, "Retrying...\n");
		snmp_pdu_free(&req);
		snmp_pdu_free(&resp);
		snmp_pdu_create(&req, SNMP_PDU_SET);
	}

	snmp_pdu_free(&req);
	snmp_pdu_free(&resp);
	snmp_tool_freeall();
	snmp_close();

	exit(0);
}
