/*-
 * Copyright (c) 2005-2006 The FreeBSD Project
 * All rights reserved.
 *
 * Author: Shteryana Shopova <syrinx@FreeBSD.org>
 *
 * Redistribution of this software and documentation and use in source and
 * binary forms, with or without modification, are permitted provided that
 * the following conditions are met:
 *
 * 1. Redistributions of source code or documentation must retain the above
 *    copyright notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *  A tool for doing a "snmp walk".
 */

#include <sys/queue.h>
#include <sys/types.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include <bsnmp/asn1.h>
#include <bsnmp/snmp.h>
#include <bsnmp/snmpclient.h>
#include "bsnmptc.h"
#include "bsnmptools.h"

/* The default tree to walk. */
static const struct asn_oid snmp_mibII_OID = {
	6 , { 1, 3, 6, 1, 2, 1 }
};

/* Help text. */
const char helptxt[] ="\
bsnmpwalk [-dhn] [-b buffersize] [-I options] [-i filelist] [-l filename]\n\
          [-o output] [-r retries] [-s [trans::][community@][server][:port]]\n\
          [-t timeout] [-v version] [OID]\n\
options: \n\
 -d		debug - dump raw Pdus\n\
 -h		print this text\n\
 -n		interpret the input as numerical OIDs rather than parse them\n\
 -b buffersize	size of receive/transmit buffer\n\
 -I options	include each of the files in the list - options are (keep\n\
		order):\n\
	cut=OID		specify an initial OID that was cut from the file to\n\
			be appended\n\
	path=pathname	specify a path where to read the files from\n\
	file=filelist	a list of comma separated files, to which the two\n\
			options above will apply\n\
 -i filelist	comma-separated filelist to read string to oid mappings\n\
 -l filename	name of the client's socket if local sockets are used\n\
 -o output	specify result output : short | verbose | quiet (default:\n\
		short)\n\
 -r retries	number of retries when sending a request (default: 3)\n\
 -s [trans::][community@][server][:port]	server specification\n\
	trans		transport type: udp | stream | dgram (default: udp)\n\
	community	community name (default: public)\n\
	server		SNMP server name or IP address\n\
	port		server port (default: 161)\n\
 -t timeout	period to wait before a request times out (default: 3 sec)\n\
 -v version	SNMP version to use : 1 | 2 (default: 2)\n\
OID		OID to start walk from - default - walk subtree rooted at\n\
		mib-2 (1.3.6.1.2.1)\n\
";

int32_t	debug_on = 0;

/* The snmp client structure from libbsnmp. */
struct snmp_client snmp_client;
struct snmp_toolinfo snmptool;

#define	DEBUG_SNMPWALK	if (debug_on) fprintf

static int32_t
snmpwalk_parse_options(int argc, char **argv)
{
	int32_t opt, count, optnum = 0;
	char opts[] = "b:dhI:i:l:no:r:s:t:v:";

	while ((opt = getopt(argc, argv, opts)) != EOF) {
	    switch (opt) {
		case 'b':
			if ((count = parse_buflen(opt, optarg)) < 0)
				return (-1);
			break;
		case 'd':
			if ((count = parse_debug(optarg)) < 0)
				return (-1);
			break;
		case 'h':
			parse_help(optarg);
			return (-1);
		case 'I':
			if ((count = parse_include(opt, optarg)) < 0)
				return (-1);
			break;
		case 'i':
			if ((count = parse_file(opt, optarg)) < 0)
				return (-1);
			break;
		case 'l':
			if ((count = parse_local_path(opt, optarg)) < 0)
				return (-1);
			break;
		case 'n':
			if ((count = parse_num_oids(optarg)) < 0)
				return (-1);
			 break;
		case 'o':
			if ((count = parse_output(opt, optarg)) < 0)
				return (-1);
			break;
		case 'r':
			if ((count = parse_retry(opt, optarg)) < 0)
				return (-1);
			break;
		case 's':
			if ((count = parse_server(opt, optarg)) < 0)
				return (-1);
			break;
		case 't':
			if ((count = parse_timeout(opt, optarg)) < 0)
				return (-1);
			break;
		case 'v':
			if ((count = parse_version(opt, optarg)) < 0)
				return (-1);
			break;
		default:
			fprintf(stderr,"Unknown option - %c\n", opt);
			fprintf(stderr, "%s", helptxt);
			return (-1);
	    }
	    optnum += count;
	}

	return (optnum);
}

static int32_t
snmpwalk_add_default(struct snmp_object *obj, char *string)
{
	asn_append_oid(&(obj->val.var), &snmp_mibII_OID);
	return (1);
}

static int32_t
snmpwalk_add_vbind(struct snmp_pdu *pdu, struct snmp_object *obj)
{
	if (pdu->nbindings > 0) {
		warnx("Too many bindings for one PDU");
		return (-1);
	}


	asn_append_oid(&(pdu->bindings[pdu->nbindings].var), &(obj->val.var));
	pdu->nbindings++;

	return (pdu->nbindings);
}

static int32_t
snmpwalk_parse_stroid(struct snmp_object *obj, char *argv)
{
	int32_t i = 0;
	char string[MAXSTR], *ptr;
	struct asn_oid in_oid;

	ptr = argv;
	while (isalpha(*ptr) || *ptr == '_' || (i != 0  && isdigit(*ptr))) {
		ptr++;
		i++;
	}

	if (i <= 0 || i >= MAXSTR)
		return (-1);

	memset(&in_oid, 0, sizeof(struct asn_oid));
	if ((ptr = snmp_parse_suboid((argv + i), &in_oid)) == NULL) {
		warnx("Invalid OID - %s", argv);
		return (-1);
	}

	strlcpy(string, argv, i + 1);
	if (snmp_lookup_oidall(obj, string) < 0) {
		warnx("No entry for %s in mapping lists", string);
		return(-1);
	}

	/*
	 * If suboid provided on command line - append it as well.
	 */

	if (in_oid.len > 0)
		asn_append_oid(&(obj->val.var), &in_oid);
	else if (*ptr == '[') {
		if ((ptr = snmp_parse_index(ptr + 1, obj)) == NULL)
			return (-1);
	}

	if (*ptr != '\0') {
		warnx("Invalid character after OID %s", argv);
		return (-1);
	}

	return (1);
}

static int32_t
snmpwalk_parse_oid(struct snmp_object *obj, char *argv)
{
	if (argv == NULL)
		return (-1);

	if (ISSET_NUMERIC) {
		if (snmp_parse_numoid(argv, &(obj->val.var)) < 0)
			return (-1);
	} else {
		if (snmpwalk_parse_stroid(obj, argv) < 0)
			return (-1);
	}

	return(1);
}

/*
 * Prepare the next GetNext/Get PDU to send.
 */
static void
snmpwalk_nextpdu_create(uint32_t op, struct asn_oid *var, struct snmp_pdu *pdu)
{
	snmp_pdu_create(pdu, op);
	asn_append_oid(&(pdu->bindings[0].var), var);
	pdu->nbindings = 1;
}

/*
 * Do a 'snmp walk' - according to command line options request for values
 * lexicographically subsequent and subrooted at a common node. Send a GetNext
 * PDU requesting the value for each next variable and print the responce. Stop
 * when a Responce PDU is received that contains the value of a variable not
 * subrooted at the variable the walk started.
 */
int
main(int argc, char **argv)
{
	int32_t oid, opt_num, outputs;
	struct asn_oid root;	/* Keep the inital oid. */
	struct snmp_pdu pdu_to_send, pdu_to_recv;

	snmptool_init(&snmptool, helptxt);
	if ((opt_num = snmpwalk_parse_options(argc, argv)) < 0) {
		snmp_tool_freeall();
		exit(0);
	}

	if (snmp_import_all() < 0) {
		snmp_tool_freeall();
		exit(0);
	}

	oid = argc - opt_num - 1;
	switch (oid) {
	    case 0:
		if (snmp_object_add(snmpwalk_add_default, NULL) < 0) {
			snmp_tool_freeall();
			fprintf(stderr, "Error setting default subtree\n");
			exit(0);
		}
		break;

	    case 1:
		/*
		 * The last command line argument will always be the OID to
		 * start the walk from.
		 */
		if ((snmp_object_add(snmpwalk_parse_oid, argv[argc - 1])) < 0) {
			snmp_tool_freeall();
			fprintf(stderr,"OID parsing error - %s\n",
			    argv[argc - 1]);
			exit(0);
		}
		break;

	    default:
		warnx("Only one OID allowed");
		snmp_tool_freeall();
		exit(0);
	}

	snmp_pdu_create(&pdu_to_send, SNMP_PDU_GETNEXT);
	if (snmp_pdu_add_bindings(NULL, snmpwalk_add_vbind, &pdu_to_send) < 0) {
		snmp_tool_freeall();
		exit(0);
	}

	if (snmp_open(NULL, NULL, NULL, NULL)) {
		warnx("Failed to open snmp session - %s", strerror (errno));
		snmp_tool_freeall();
		exit(-1);
	}

	/* Remember the root where the walk started from. */
	memset(&root, 0, sizeof(struct asn_oid));
	asn_append_oid(&root, &(pdu_to_send.bindings[0].var));

	outputs = 0;
	while (snmp_dialog(&pdu_to_send, &pdu_to_recv) >= 0) {
		if ((snmp_parse_resp(&pdu_to_recv, &pdu_to_send)) < 0) {
			snmp_output_err_resp(&pdu_to_recv);
			snmp_pdu_free(&pdu_to_recv);
			outputs = -1;
			break;
		}

		if (!(asn_is_suboid(&root, &(pdu_to_recv.bindings[0].var)))) {
			snmp_pdu_free(&pdu_to_recv);
			break;
		}
		snmp_output_resp(&pdu_to_recv);
		outputs++;
		snmp_pdu_free(&pdu_to_recv);

		snmpwalk_nextpdu_create(SNMP_PDU_GETNEXT,
		    &(pdu_to_recv.bindings[0].var), &pdu_to_send);
	}

	/* Just a case our root was a leaf. */
	if (outputs == 0) {
		snmpwalk_nextpdu_create(SNMP_PDU_GET, &root, &pdu_to_send);
		if (snmp_dialog(&pdu_to_send, &pdu_to_recv) == SNMP_CODE_OK) {
			if (snmp_parse_resp(&pdu_to_recv,&pdu_to_send) < 0)
				snmp_output_err_resp(&pdu_to_recv);
			else
				snmp_output_resp(&(pdu_to_recv));
			snmp_pdu_free(&pdu_to_recv);
		} else
			warnx("Snmp dialog - %s", strerror (errno));
	}

	snmp_tool_freeall();
	snmp_close();
	exit(0);
}
